# Problem Statement:
# ------------------
# Solution of the mixed system on a unit cube domain (2D space, time in 3rd dimension):
# div(\sigma) + b = 0 on a unit cube. Body force b=0 in this problem
#
# BCs:
# ----
# Dirichlet: ux = 0 on left face
#            uy = 0 on nodes along y=0.5 and on "left" face (x=0)
# Neumann  : \sigma \cdot n = force on right edge
# 
# The linear elastodynamic equation is solved using the FOSLS variational form
# 
# Analytical Solution:
# ux_ex = (1+nu)*(1-nu)/E * Tx * x
# uy_ex = -(1+nu)*nu/E * Tx * (y-0.5) 
# ***************************************************
from ngsolve import *
from netgen.csg import *
import netgen.gui

# ***************************************************
# Generate geometry for a unit cube
# ***************************************************
def Geom_Cube1(xb,yb,Tb, xe,ye,Te):
   #geo_cube = CSGeometry()
   left  = Plane (Pnt(xb,yb,Tb), Vec(-1,0,0) ).bc("left")
   right = Plane (Pnt(xe,ye,Te), Vec( 1,0,0) ).bc("right")
   front = Plane (Pnt(xb,yb,Tb), Vec(0,-1,0) ).bc("front")
   back  = Plane (Pnt(xe,ye,Te), Vec(0, 1,0) )
   bot   = Plane (Pnt(xb,yb,Tb), Vec(0,0,-1) ).bc("bottom")
   top   = Plane (Pnt(xe,ye,Te), Vec(0,0, 1) ).bc("top")
   
   cube = left * right * front * back * bot * top
   #geo_cube.Add (cube)
   #geo.Draw()
   return cube


def Geom_Cube2(xb,yb,Tb, xe,ye,Te):
   #geo_cube = CSGeometry()
   left  = Plane (Pnt(xb,yb,Tb), Vec(-1,0,0) ).bc("left")
   right = Plane (Pnt(xe,ye,Te), Vec( 1,0,0) ).bc("right")
   front = Plane (Pnt(xb,yb,Tb), Vec(0,-1,0) )
   back  = Plane (Pnt(xe,ye,Te), Vec(0, 1,0) ).bc("back")
   bot   = Plane (Pnt(xb,yb,Tb), Vec(0,0,-1) ).bc("bottom")
   top   = Plane (Pnt(xe,ye,Te), Vec(0,0, 1) ).bc("top")
   
   cube = left * right * front * back * bot * top
   #geo_cube.Add (cube)
   #geo.Draw()
   return cube

# The unit cube is generated from 2 parts: 0<y<0.5 and 0.5<=y<1.0 so as to enforce mesh nodes along y=0.5. This is done to apply Dirichlet BCs uy=0 at y=0.5 on the left face.
cube1 = Geom_Cube1(0,0,0, 1.0,0.5,1.0)
cube2 = Geom_Cube2(0,0.5,0, 1.0, 1.0, 1.0)
cube  = cube1 + cube2

geo_cube = CSGeometry()
geo_cube.Add (cube)
maxh_ele = 0.5
ngmesh = geo_cube.GenerateMesh(maxh=maxh_ele)
#ngmesh.Save("cube.vol")
mesh=Mesh(ngmesh)

# Define element order
ele_order_u   = 2
ele_order_sig = 1

##   Elasticity parameters
E   = 207.0e3 # Elastic modulus
nu  = 0.35    # Poisson's ratio
rho = 1.0     # Mass density
#
Tx = z
Ty = 0.0

# Analytical Solution
ux_ex = (1+nu)*(1-nu)/E * Tx * x
uy_ex = -(1+nu)*nu/E * Tx * (y-0.5)
#
duxdt0 = (1+nu)*(1-nu)/E * x
duydt0 = -(1+nu)*nu/E * (y-0.5)

# Traction loading specification (Neumann BC)
#force = CoefficientFunction( (100.0,0.0) )
force1 = CoefficientFunction([ Tx if bc=="right" else rho*duxdt0 if bc=="bottom" else 0 for bc in mesh.GetBoundaries()])
force2 = CoefficientFunction([ rho*duydt0 if bc=="bottom" else 0 for bc in mesh.GetBoundaries()])
#
##   2D Approximation
model = "plane strain" # same as in 3D
#model = "plane_stress"

## Alternate Elasticity parameters (dependent on E, nu and 2D approximation)
mu = E/2/(1+nu)
lmbda = E*nu/(1+nu)/(1-2*nu)
if model == "plane_stress":
    lmbda = 2*mu*lmbda/(lmbda+2*mu)

########################################################
# some utilities ...
########################################################
# 1. Count DOFs with no Dirichlet BCs
def FreeDofCount(fespace):
  
    freedofs=fespace.FreeDofs()
    cntfr=0
    for i in range(fespace.ndof):
        if freedofs[i]==1:
            cntfr += 1
    return cntfr
#    
def IdentityCF(dim):
    return CoefficientFunction( tuple( [1 if i==j else 0 for i in range(dim) for j in range(dim)]), dims=(dim,dim) )

def Trace(mat):
    return sum( [mat[i,i] for i in range(mat.dims[0]) ])

# Define Strain Tensor
def epsSp(ux, uy):
    #du = 0.5*(u.Deriv() + u.Deriv().trans) # NOTE: row .. component, col .. derivative
    eps11 = grad(ux)[0]
    eps12 = 0.5*(grad(ux)[1] + grad(uy)[0])
    eps21 = eps12
    eps22 = grad(uy)[1]
    return CoefficientFunction((eps11, eps12, eps21, eps22), dims=(2,2))

# Compute time derivative of ith component of u    
def dudt(u):
    return grad(u)[2] # 2 corresponds to time axis

# Build the stress tensor from stress-vectors    
def sigma(s1, s2):
    sig11 = s1[0]
    sig12 = s1[1]
    sig21 = s2[0]
    sig22 = s2[1]
    return CoefficientFunction((sig11, sig12, sig21, sig22), dims=(2,2))

# Build the stress tensor from stress-vectors 
def tau(s1, s2):
    sig11 = s1[0]
    sig12 = s1[1]
    sig21 = s2[0]
    sig22 = s2[1]
    return CoefficientFunction((sig11, sig12, sig21, sig22), dims=(2,2))    

# Product of Kronecker deltas with same indices
def ddSp():
    return mesh.dim-1

# Define spatial divergence of a vector field in space-time
def divSp(ux, uy):
    #return Trace(epsSp(u))
    dux = grad(ux)[0]
    duy = grad(uy)[1]
    return  (dux+duy)
    
########################################################    
# Define H1 space for displacement componenets
fesux = H1(mesh, order=ele_order_u, dirichlet="left|bottom")
fesuy = H1(mesh, order=ele_order_u, dirichlet="bottom")
#fesu = VectorH1(mesh, order=ele_order, dirichlet="bottom")

# Define HDiv space of functions for each row of the stress tensor
fes_sig1 = HDiv(mesh, order=ele_order_sig, dirichlet="bottom|right|front|back")
fes_sig2 = HDiv(mesh, order=ele_order_sig,  dirichlet="bottom|right|left|front|back")

# Compound Space
fesm = FESpace([fesux, fesuy, fes_sig1, fes_sig2])

def y_bc():
# Dirichlet boundary conditions on displacement field
    freedofs = fesm.FreeDofs()
    tol = 1.0E-14
    
    #Apply uy=0 on nodes along y=0.5 and on "left" face (x=0)
    for ve in mesh.vertices:
        if ve.point[0] < tol and (ve.point[1] > 0.5-tol and ve.point[1] < 0.5+tol):
             for dof in fesux.GetDofNrs(ve):
                 freedofs.Clear(fesux.ndof + dof)

# Apply boundary condition on 'uy'
y_bc()
                 
# Test and Trial Functions from the defined spaces
ux, uy, sig1, sig2 = fesm.TrialFunction()
vx, vy, tau1, tau2 = fesm.TestFunction()

sigT = sigma(sig1, sig2)
tauT = tau(tau1, tau2)

# Bilinear Form
am = BilinearForm(fesm)
# Part I: Stress terms in linear momentum balance
am += (div(sig1)*div(tau1) + div(sig2)*div(tau2))*dx

# Part Ia: Inertial terms in linear momentum balance
am += (sig1[2]+rho*dudt(ux))*tau1[2]*dx
am += (sig2[2]+rho*dudt(uy))*tau2[2]*dx
am += (sig1[2]+rho*dudt(ux))*rho*dudt(vx)*dx
am += (sig2[2]+rho*dudt(uy))*rho*dudt(vy)*dx

# Part II: Terms from constitutive relation
am += -lmbda*(divSp(vx,vy)*Trace(sigT) + divSp(ux,uy)*Trace(tauT))*dx
am += -2*mu*(InnerProduct(sigT,epsSp(vx,vy)) + InnerProduct(tauT,epsSp(ux,uy)))*dx
am += InnerProduct(sigT,tauT)*dx
am += lmbda*lmbda*divSp(ux,uy)*divSp(vx,vy)*ddSp()*dx
am += 4*mu*lmbda*divSp(ux,uy)*divSp(vx,vy)*dx
am += 4*mu*mu*InnerProduct(epsSp(ux,uy),epsSp(vx,vy))*dx

fm = LinearForm(fesm)  # No body force

# code crashes if preconditioner type="multigrid"
preC = Preconditioner(am, type="bddc") 

gfesm = GridFunction(fesm, name="gfmixed")
gfux, gfuy, gfsig1, gfsig2 = gfesm.components

def SolveBVP():
    # Apply essential bc on 'y' dof - this cannot be specified as part of the function space
    y_bc()
    
    # Assemble bilinear and linear forms
    am.Assemble()
    fm.Assemble()
    
    # Traction BCs: Neumann bc of origial problem - essential bc in mixed form
    normal = specialcf.normal(mesh.dim)
    #BC on sigma in Hdiv
    gfsig1.Set(force1*normal, BND)
    gfsig2.Set(force2*normal, BND)

    res = fm.vec.CreateVector()
    res[:] = 0.0
    res.data = fm.vec - am.mat*gfesm.vec
    
    Kinv = CGSolver(am.mat, preC.mat, maxsteps=10000)
    gfesm.vec.data += Kinv * res
    
    #solvers.BVP(bf=am_F, lf=fm_F, gf=gfesm_F)
    Redraw (blocking=True)
    
#Compute FOSLS residual
l = []
def CalcRes():

    # Compute L2 norms of each part of FOSLS functional
    FOS1a = div(gfsig1)
    FOS1b = div(gfsig2)
    FOS2T = sigma(gfsig1,gfsig2) - lmbda*divSp(gfux,gfuy)*IdentityCF(mesh.dim-1) - 2*mu*epsSp(gfux, gfuy)
    #FOS2 = FOS2T[0,0]+FOS2T[0,1]+FOS2T[1,0]+FOS2T[1,1]
    FOS3a = gfsig1[2]+rho*dudt(gfux)
    FOS3b = gfsig2[2]+rho*dudt(gfuy)

    elres_FOS1 = Integrate(FOS1a*FOS1a+FOS1b*FOS1b, mesh, VOL, element_wise=True)
    elres_FOS2 = Integrate(FOS2T[0,0]*FOS2T[0,0]+FOS2T[0,1]*FOS2T[0,1]+FOS2T[1,0]*FOS2T[1,0]+FOS2T[1,1]*FOS2T[1,1],mesh,VOL,element_wise=True)
    elres_FOS3 = Integrate(FOS3a*FOS3a+FOS3b*FOS3b,mesh,VOL,element_wise=True)
    elres = elres_FOS1 + elres_FOS2 + elres_FOS3
    max_elres = sqrt(max(elres))
    
    res_FOS1_Vol  = sqrt(sum(elres_FOS1))
    res_FOS2_Vol  = sqrt(sum(elres_FOS2))
    res_FOS3_Vol  = sqrt(sum(elres_FOS3))
    
    res_tot_Vol = res_FOS1_Vol + res_FOS2_Vol + res_FOS3_Vol
    
    l.append ( (fesm.ndof, FreeDofCount(fesm), res_FOS1_Vol, res_FOS2_Vol, res_FOS3_Vol, res_tot_Vol) )
    print ("max element residual = ", max_elres)

    for el in mesh.Elements():
        mesh.SetRefinementFlag(el, sqrt(elres[el.nr]) >= 0.8*max_elres)
        
    for el in mesh.Elements(BND):
        mesh.SetRefinementFlag(el, False)

def Update_Sp_GF():
    # Update Function Spaces
    fesux.Update()
    fesuy.Update()
    fes_sig1.Update()
    fes_sig2.Update()
#    fes_L.Update()
    fesm.Update()

    # Update corresponding grid functions
    gfux.Update()
    gfuy.Update()
    gfsig1.Update()
    gfsig2.Update()
#    gfL.Update()
    gfesm.Update()

    # Update preconditioner
    #preC.Update()

def Draw_res():
    Draw(gfsig1, mesh, "sigma1")
    Draw(gfsig2, mesh, "sigma2")
    Draw(gfux, mesh, "ux-mixed")
    Draw(gfuy, mesh, "uy-mixed")

isolve = 0
import numpy as np
with TaskManager():
     #while fesm.ndof < 2000000: 
     while isolve<2:
        SolveBVP()
        if isolve==0:
          Draw_res()
          
        isolve += 1
        print("Number of solves completed:    ", isolve)
        print("Total no. of DOFs:    ", fesm.ndof)
        print("No. of free DOFs :    ", FreeDofCount(fesm))
        CalcRes()
        mesh.Refine()
        #mesh.ngmesh.Save("mesh1.vol")
        Update_Sp_GF()

SolveBVP()
isolve += 1
print("Number of solves completed:    ", isolve)
print("Total no. of DOFs:    ", fesm.ndof)
print("No. of free DOFs :    ", FreeDofCount(fesm))
#vtkout(isolve)
CalcRes()

        
               